# 1 "<stdin>"
# 1 "<built-in>"
# 1 "<command line>"
# 1 "<stdin>"
program FWI

  use sep
  use tomo_mod

  implicit none

  include "fftw3.f"

  integer :: nt, vel_nz, vel_nx, wfld_j, dev_num, sou_z, rec_z
  integer :: it, iz, ix, ishot, counter
  integer :: sou_nx, sou_ox, sou_dx, rec_nx, rec_ox, rec_dx
  integer :: n1s, n2s, n3s, crop1, crop2, fat
  real :: dt, vel_dz, vel_oz, vel_dx, vel_ox

  real :: rho, itry, ntry
  real, dimension(:), allocatable :: wavelet, rands
  real, dimension(:,:), allocatable :: vel0, injected, wt
  real, dimension(:,:,:), allocatable :: sou_wfld, rec_wfld
  complex, dimension(:,:,:), allocatable :: csou_wfld, crec_wfld
  real, dimension(:,:,:,:), allocatable :: refl

  integer :: sou_x, rec_x, irec
  integer :: nlagz, nlagx, lagz, lagx, ilagz, ilagx
  integer :: lbz1, ubz1, lbx1, ubx1
  integer :: lbz2, ubz2, lbx2, ubx2

  integer, dimension(8) :: timer0, timer1
  real :: timer_sum=0.

  complex, dimension(:), allocatable :: in1, out1

  integer*8 plan1f

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------
  ! Read background velocity geometry
  call from_history("n1", vel_nz)
  call from_history("o1", vel_oz)
  call from_history("d1", vel_dz)
  call from_history("n2", vel_nx)
  call from_history("o2", vel_ox)
  call from_history("d2", vel_dx)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  ! Read recording time
  call from_param("nt", nt)
  call from_aux("wavfile", "d1", dt)

  ! Read in Density
  call from_param("rho", rho)

  call from_param("wfld_j", wfld_j)
  call from_param("dev_num", dev_num, 0)
  call from_param("crop1", crop1)
  call from_param("crop2", crop2)

  call from_param("nlagz", nlagz)
  call from_param("nlagx", nlagx)

  call from_param("ntry", ntry)

  fat=5

  n3s = ceiling(1.0*nt/wfld_j)
  n1s = vel_nz-2*crop1
  n2s = vel_nx-2*crop2

  ! Allocation
  allocate( wavelet(nt))
  allocate( injected(nt, vel_nx))
  allocate( vel0(vel_nz, vel_nx), wt(vel_nz, vel_nx))
  allocate( sou_wfld(n1s, n2s, n3s), rec_wfld(n1s, n2s, n3s))
  allocate( in1(n3s), out1(n3s))
  allocate( csou_wfld(n3s, n1s, n2s), crec_wfld(n3s, n1s, n2s))
  allocate( refl(n1s, n2s, 2*nlagz+1, 2*nlagx+1))
  allocate( rands(rec_nx))

  ! Read in the background velocity
  call sep_read(vel0)
  vel0 = 1./sqrt(vel0)

  ! Read in the source wavefields
  call sep_read(wavelet, "wavfile")

  ! Scale the source wavefields by rho
  wavelet = wavelet * rho

  call SecDer1d(wavelet, wavelet)

  call to_history("n1", n1s)
  call to_history("o1", vel_oz+crop1*vel_dz)
  call to_history("d1", vel_dz)
  call to_history("n2", n2s)
  call to_history("o2", vel_ox+crop2*vel_dx)
  call to_history("d2", vel_dx)
  call to_history("n3", 2*nlagz+1)
  call to_history("o3", -1.*nlagz*vel_dz)
  call to_history("d3", vel_dz)
  call to_history("n4", 2*nlagx+1)
  call to_history("o4", -1.*nlagx*vel_dx)
  call to_history("d4", vel_dx)
  call to_history("n5", ntry)
  call to_history("o5", 0.)
  call to_history("d5", 1.)

  call sfftw_plan_dft_1d(plan1f,n3s,in1,out1,FFTW_FORWARD,FFTW_MEASURE)

  call c_init(nt, dt, vel_nz, vel_nx, vel_dz, vel_dx, dev_num)
  call c_init_prop(n1s, n2s, crop1, crop2, wfld_j)

  call c_vel_update(vel0*dt/vel_dx/4., vel0*vel0)
  wt = 0.
  do ix=1+fat,crop2
    do iz=1+fat,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float((crop2-ix)+1)/float(crop2)*.05*((vel0(iz,ix)*dt/vel_dx)/.333)
    enddo
  enddo
  do ix=vel_nx-crop2+1,vel_nx-fat
    do iz=1+fat,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float(ix-(vel_nx-crop2+1)+1)/float(crop2)*.05*((vel0(iz,ix)*dt/vel_dx)/.333)
    enddo
  enddo
  do ix=1+fat,vel_nx-fat
    do iz=1+fat,crop1
      wt(iz,ix) = wt(iz,ix) + float((crop1-iz)+1)/float(crop1)*.05*((vel0(iz,ix)*dt/vel_dz)/.333)
    enddo
  enddo
  do ix=1+fat,vel_nx-fat
    do iz=vel_nz-crop1+1,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float(iz-(vel_nz-crop1+1)+1)/float(crop1)*.05*((vel0(iz,ix)*dt/vel_dz)/.333)
    enddo
  enddo
  call c_wt_update(wt*4.)

  refl = 0.

  do itry=1,ntry

    do ishot=1,sou_nx
      sou_x = sou_ox+1+(ishot-1)*sou_dx

      injected = 0.
      do it=1,nt
        injected(it,sou_x) = wavelet(it)
      enddo
      sou_wfld = 0.
      call c_prop(sou_z, injected, sou_wfld, 0)

      do
        call random_number(rands)
        do irec=1,rec_nx
          if(rands(irec) < 0.5) then
            rands(irec) = -1.
          else
            rands(irec) = 1.
          endif
        enddo
        if(abs(sum(rands)) <=1) exit
      enddo

      injected = 0.
      do irec=1,rec_nx
        rec_x = rec_ox+1+(irec-1)*rec_dx
        do it=1,nt
          injected(it,rec_x) = wavelet(it)*rands(irec)
        enddo
      enddo ! end of rec loop

      rec_wfld = 0.
      call c_prop(rec_z, injected, rec_wfld, 0)

      csou_wfld = 0.
      do ix=1,n2s
        do iz=1,n1s
          in1 = sou_wfld(iz,ix,:)
          call sfftw_execute_dft(plan1f, in1, out1)
          csou_wfld(:,iz,ix) = out1
        enddo
      enddo

      crec_wfld = 0.
      do ix=1,n2s
        do iz=1,n1s
          in1 = rec_wfld(iz,ix,:)
          call sfftw_execute_dft(plan1f, in1, out1)
          crec_wfld(:,iz,ix) = out1
        enddo
      enddo

      csou_wfld = csou_wfld*crec_wfld

        do ilagx=1,2*nlagx+1
        lagx = ilagx-1-nlagx
        if(lagx > 0) then
          lbx1 = 1
          ubx1 = n2s-lagx
          lbx2 = 1+lagx
          ubx2 = n2s
        else
          lbx1 = 1-lagx
          ubx1 = n2s
          lbx2 = 1
          ubx2 = n2s+lagx
        endif
        do ilagz=1,2*nlagz+1
          lagz = ilagz-1-nlagz
          if(lagz > 0) then
            lbz1 = 1
            ubz1 = n1s-lagz
            lbz2 = 1+lagz
            ubz2 = n1s
          else
            lbz1 = 1-lagz
            ubz1 = n1s
            lbz2 = 1
            ubz2 = n1s+lagz
          endif

          refl(lbz1:ubz1,lbx1:ubx1,ilagz,ilagx) = refl(lbz1:ubz1,lbx1:ubx1,ilagz,ilagx) + sum(csou_wfld(:,lbz1:ubz1,lbx1:ubx1)*conjg(csou_wfld(:,lbz2:ubz2,lbx2:ubx2)),1)

        enddo
      enddo

    enddo ! end of shot loop

    call sep_write(refl/itry)
  enddo


!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
