# 1 "<stdin>"
# 1 "<built-in>"
# 1 "<command line>"
# 1 "<stdin>"
module cdstep3_mod
  implicit none
  integer, private :: n1m, n2m, n1d, n2d, n3d
  real, dimension (:,:), allocatable, private :: s
  real, dimension (:,:,:), allocatable, private :: ss

  contains

  subroutine cdstep_init(n1m_in, n2m_in, n1d_in, n2d_in, n3d_in)
    integer :: n1m_in, n2m_in, n1d_in, n2d_in, n3d_in
    n1m = n1m_in
    n2m = n2m_in
    n1d = n1d_in
    n2d = n2d_in
    n3d = n3d_in
  end subroutine

  integer function cdstep( forget, x, g, rr, gg)
    real, dimension(n1m, n2m) :: x, g
    real, dimension(n1d, n2d, n3d) :: rr, gg
    logical :: forget
    double precision :: sds, gdg, gds, determ, gdr, sdr, alfa, beta
    if ( .not. allocated (s)) then
      forget = .true.
      allocate ( s (n1m, n2m))
      allocate (ss (n1d, n2d, n3d))
    end if
    if ( forget) then
      s = 0.
      ss = 0.
      beta = 0.d0 ! steepest descent
      if ( sum(gg*gg) .eq. 0 ) then
        call erexit('cdstep: grad vanishes identically')
      end if
      alfa = - sum( dprod( gg, rr)) / sum( dprod( gg, gg))
    else
      gdg = sum( dprod( gg, gg))
      ! search plane by solving 2-by-2
      sds = sum( dprod( ss, ss))
      ! G . (R - G*alfa - S*beta) = 0
      gds = sum( dprod( gg, ss))
      ! S . (R - G*alfa - S*beta) = 0
      if ( gdg.eq.0. .or. sds.eq.0.) then
        cdstep = 1
        return
      end if
      determ = gdg * sds * max (1.d0 - (gds/gdg)*(gds/sds), 1.d-12)
      gdr = - sum( dprod( gg, rr))
      sdr = - sum( dprod( ss, rr))
      alfa = ( sds * gdr - gds * sdr ) / determ
      beta = (-gds * gdr + gdg * sdr ) / determ
    end if
    write(0,*) "alfa, beta", alfa, beta
    s = alfa * g + beta * s ! update solution step
    ss = alfa * gg + beta * ss ! update residual step
    x = x + s ! update solution
    rr = rr + ss ! update residual
    forget = .false.
    cdstep = 0
  end function

  subroutine cdstep_close ( )
    if ( allocated( s)) then
      deallocate( s, ss)
    end if
  end subroutine
end module
