program Forward

  use sep
  use tomo_mod

  implicit none

  integer                             :: nt, vel_nz, vel_nx, dev_num, sou_z, rec_z, i, it, ix, iz
  integer                             :: sou_nx, sou_ox, sou_dx, rec_nx, rec_ox, rec_dx
  integer                             :: crop1, crop2, fat
  real                                :: dt, vel_dz, vel_oz, vel_dx, vel_ox

  real                                :: rho
  real, dimension(:), allocatable     :: wavelet
  real, dimension(:,:), allocatable   :: vel0, refl, wavelet2, data, wt
!  real, dimension(:,:,:), allocatable :: wfld
!  integer                             :: n1s, n2s, n3s, wfld_j

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  ! Read background velocity geometry
  call from_history("n1", vel_nz)
  call from_history("o1", vel_oz)
  call from_history("d1", vel_dz)
  call from_history("n2", vel_nx)
  call from_history("o2", vel_ox)
  call from_history("d2", vel_dx)

  if(mod(vel_nz-10,16) .ne. 0) call erexit("(n1-10)/20 is not an integer")
  if(mod(vel_nx-10,16) .ne. 0) call erexit("(n2-10)/20 is not an integer")

  ! Read wavelet information
  call from_aux("wavfile", "d1", dt)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  ! Read recording time
  call from_param("nt", nt)

  ! Read in Density
  call from_param('rho', rho, 1000.0)
  call from_param("crop1",crop1)
  call from_param("crop2",crop2)
  call from_param("dev_num",dev_num)

  if(sou_nx .eq. 0) sou_nx=vel_nx
  if(rec_nx .eq. 0) rec_nx=vel_nx

  fat=5

!  crop1 = 0
!  crop2 = 0
!  wfld_j = 2

!  n1s = vel_nz-2*crop1
!  n2s = vel_nx-2*crop2
!  n3s = ceiling(1.0*nt/wfld_j)

  ! Allocation
  allocate( wavelet(nt))
  allocate( vel0(vel_nz, vel_nx), refl(vel_nz, vel_nx), wavelet2(nt, vel_nx), data(nt, rec_nx), wt(vel_nz, vel_nx))
!  allocate( wfld(n1s, n2s, n3s))

  call to_history("n1", nt)
  call to_history("o1", 0.0)
  call to_history("d1", dt)
  call to_history("n2", rec_nx)
  call to_history("o2", vel_dx*rec_ox+vel_ox)
  call to_history("d2", vel_dx)
  call to_history("n3", sou_nx)
  call to_history("o3", vel_dx*sou_ox+vel_ox)
  call to_history("d3", vel_dx*sou_dx)
  call to_history("label1", "Time(s)")
  call to_history("label2", "Location(km)")
  call to_history("label3", " ")

!  call to_history("n1", n1s)
!  call to_history("n2", n2s)
!  call to_history("n3", n3s)

  ! Read in the background velocity
  call sep_read(vel0)
  vel0 = 1./sqrt(vel0)

  call sep_read(refl, "refl")

  ! Read in the source wavefields
  call sep_read(wavelet, "wavfile")

  ! Scale the source wavefields by rho
  wavelet = wavelet * rho
  call SecDer1d(wavelet, wavelet)

  wt = 0.

!  crop1 = 50
!  crop2 = 50

  do ix=1+fat,crop2
    do iz=1+fat,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float((crop2-ix)+1)/float(crop2)*.05*((vel0(iz,ix)*dt/vel_dx)/.333)
    enddo
  enddo

  do ix=vel_nx-crop2+1,vel_nx-fat
    do iz=1+fat,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float(ix-(vel_nx-crop2+1)+1)/float(crop2)*.05*((vel0(iz,ix)*dt/vel_dx)/.333)
    enddo
  enddo

  do ix=1+fat,vel_nx-fat
    do iz=1+fat,crop1
      wt(iz,ix) = wt(iz,ix) + float((crop1-iz)+1)/float(crop1)*.05*((vel0(iz,ix)*dt/vel_dz)/.333)
    enddo
  enddo

  do ix=1+fat,vel_nx-fat
    do iz=vel_nz-crop1+1,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float(iz-(vel_nz-crop1+1)+1)/float(crop1)*.05*((vel0(iz,ix)*dt/vel_dz)/.333)
    enddo
  enddo

!  crop1 = 0
!  crop2 = 0

!  call sep_write(wt,"wt.H")
!  call to_history("n1",vel_nz,"wt.H")
!  call to_history("n2",vel_nx,"wt.H")

  call c_init(nt, dt, vel_nz, vel_nx, vel_dz, vel_dx, dev_num)
!  call c_init_prop(n1s, n2s, crop1, crop2, wfld_j)
  call c_init_model(nt, rec_nx)
  call c_init_model2(vel_nz, vel_nx)
  call c_vel_update(vel0*dt/vel_dx/4., vel0*vel0)
  call c_wt_update(wt*4.)

  do i=sou_ox, (sou_nx-1)*sou_dx+sou_ox, sou_dx
!  do i=sou_ox, sou_ox

    wavelet2 = 0.
    wavelet2(:,i+1) = wavelet(:)
    data = 0.

!    call c_prop(sou_z, wavelet2, wfld, 0)
!    call c_model(sou_z, wavelet2, rec_z, rec_ox, data)
    call c_model2(sou_z, wavelet2, refl, rec_z, rec_ox, data)
    call sep_write(data)
!    call sep_write(wfld)

  end do

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
