program LSRTM

  use sep
  use tomo_mod

  implicit none

  include "fftw3.f"

  integer                             :: nt, vel_nz, vel_nx, wfld_j, dev_num, sou_z, rec_z
  integer                             :: it, iz, ix, iter, niter, ishot, counter
  integer                             :: sou_nx, sou_ox, sou_dx, rec_nx, rec_ox, rec_dx
  integer                             :: n1s, n2s, n3s, crop1, crop2, stat, fat
  real                                :: dt, vel_dz, vel_oz, vel_dx, vel_ox

  real                                :: rho
  real, dimension(:), allocatable     :: wavelet, rands_sou
  real, dimension(:,:), allocatable   :: vel0, refl, injected, gradient, wt
  real, dimension(:,:), allocatable   :: data_diff, gradient_data, hess
  real, dimension(:,:,:), allocatable :: sou_wfld, rec_wfld, data_obs

  integer                             :: sou_x, rec_x, irec
  logical                             :: forget
  double precision                    :: alfa

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  complex, dimension(:), allocatable  :: in1, out1
  integer*8 plan1f

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------
  ! Read background velocity geometry
  call from_history("n1", vel_nz)
  call from_history("o1", vel_oz)
  call from_history("d1", vel_dz)
  call from_history("n2", vel_nx)
  call from_history("o2", vel_ox)
  call from_history("d2", vel_dx)

  ! Read shots geometry
  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  ! Read receivers geometry
  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  ! Read recording time
  call from_aux("data_obs", "n1", nt)
  call from_aux("data_obs", "d1", dt)

  ! Read in Density
  call from_param("rho", rho)

  call from_param("wfld_j", wfld_j)
  call from_param('niter', niter, 2)
  call from_param("dev_num", dev_num, 0)
  call from_param("crop1", crop1)
  call from_param("crop2", crop2)

  fat=5

  n3s = ceiling(1.0*nt/wfld_j)
  n1s = vel_nz-2*crop1
  n2s = vel_nx-2*crop2

  ! Allocation
  allocate( wavelet(nt))
  allocate( injected(nt, vel_nx), gradient(vel_nz, vel_nx))
  allocate( vel0(vel_nz, vel_nx), refl(vel_nz, vel_nx), wt(vel_nz, vel_nx))
  allocate( sou_wfld(n1s, n2s, n3s), rec_wfld(n1s, n2s, n3s))
  allocate( data_diff(nt, rec_nx), gradient_data(nt, rec_nx))
  allocate( data_obs(nt, rec_nx, sou_nx))
  allocate( rands_sou(sou_nx))
  allocate( in1(n3s), out1(n3s))
  allocate( hess(vel_nz, vel_nx))

  ! Read in the background velocity
  call sep_read(vel0)
  vel0 = 1./sqrt(vel0)

  ! Read in the observed data
  call sep_read(data_obs, "data_obs")

  ! Read in the source wavefields
  call sep_read(wavelet, "wavfile")

  ! Scale the source wavefields by rho
  wavelet = wavelet * rho

  call SecDer1d(wavelet, wavelet)

  call to_history("n1", vel_nz)
  call to_history("o1", vel_oz)
  call to_history("d1", vel_dz)
  call to_history("n2", vel_nx)
  call to_history("o2", vel_ox)
  call to_history("d2", vel_dx)
  call to_history("o3", 0.0)
  call to_history("d3", 1.0)

  call sfftw_plan_dft_1d(plan1f,n3s,in1,out1,FFTW_FORWARD,FFTW_MEASURE)

  call c_init(nt, dt, vel_nz, vel_nx, vel_dz, vel_dx, dev_num)
  call c_init_prop(n1s, n2s, crop1, crop2, wfld_j)
  call c_init_model(nt, rec_nx)
  call c_init_model2(vel_nz, vel_nx)

  call c_vel_update(vel0*dt/vel_dx/4., vel0*vel0)
  wt = 0.
  do ix=1+fat,crop2
    do iz=1+fat,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float((crop2-ix)+1)/float(crop2)*.05*((vel0(iz,ix)*dt/vel_dx)/.333)
    enddo
  enddo
  do ix=vel_nx-crop2+1,vel_nx-fat
    do iz=1+fat,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float(ix-(vel_nx-crop2+1)+1)/float(crop2)*.05*((vel0(iz,ix)*dt/vel_dx)/.333)
    enddo
  enddo
  do ix=1+fat,vel_nx-fat
    do iz=1+fat,crop1
      wt(iz,ix) = wt(iz,ix) + float((crop1-iz)+1)/float(crop1)*.05*((vel0(iz,ix)*dt/vel_dz)/.333)
    enddo
  enddo
  do ix=1+fat,vel_nx-fat
    do iz=vel_nz-crop1+1,vel_nz-fat
      wt(iz,ix) = wt(iz,ix) + float(iz-(vel_nz-crop1+1)+1)/float(crop1)*.05*((vel0(iz,ix)*dt/vel_dz)/.333)
    enddo
  enddo
  call c_wt_update(wt*4.)

  refl = 0.
  hess = 0.
  counter = 0

  do iter=1,niter

    gradient = 0.

    rands_sou = 0.
    do
      call random_number(rands_sou)
      do ishot=1,sou_nx
        if(rands_sou(ishot) < 0.5) then 
           rands_sou(ishot) = -1.
        else
          rands_sou(ishot) = 1.
        endif
      enddo
      if(abs(sum(rands_sou)) <=1) exit
    enddo

    injected = 0.
    do ishot=1,sou_nx
      injected(:,sou_ox+1+(ishot-1)*sou_dx) = rands_sou(ishot)*wavelet
    end do ! end of shot loop
    data_diff = 0.
    call c_model2(sou_z, injected, refl, rec_z, rec_ox, data_diff)

    do ishot=1,sou_nx
      data_diff = data_diff - rands_sou(ishot)*data_obs(:,:,ishot)
    end do ! end of shot loop

    sou_wfld = 0.
    call c_prop(sou_z, injected, sou_wfld, 0)

    injected = 0.
    do ix=1,rec_nx
      injected(:,rec_ox+ix) = data_diff(:,ix)
    end do
    rec_wfld = 0.
    call c_prop(rec_z, injected, rec_wfld, n3s-1)

    gradient(1+crop1:vel_nz-crop1,1+crop2:vel_nx-crop2) = - sum(sou_wfld*rec_wfld,3)

    do ix=1,n2s
      do iz=1,n1s
        in1 = sou_wfld(iz,ix,:)
        call sfftw_execute_dft(plan1f, in1, out1)
        out1 = out1*out1
        hess(iz+crop1,ix+crop2) = hess(iz+crop1,ix+crop2) + sum(out1*conjg(out1))
      enddo
    enddo

    do ix=1,crop2
      hess(:,ix) = hess(:,1+crop2)
    enddo
    do ix=vel_nx-crop2+1,vel_nx
      hess(:,ix) = hess(:,vel_nx-crop2)
    enddo
    do iz=1,crop1
      hess(iz,:) = hess(1+crop1,:)
    enddo
    do iz=vel_nz-crop1+1,vel_nz
      hess(iz,:) = hess(vel_nz-crop1,:)
    enddo

    if(minval(hess) .eq. 0.) hess = hess + 1.

    gradient = gradient/hess*(1.+counter)

    do ix=1,crop2
      gradient(:,ix) = gradient(:,1+crop2)
    enddo
    do ix=vel_nx-crop2+1,vel_nx
      gradient(:,ix) = gradient(:,vel_nx-crop2)
    enddo
    do iz=1,crop1
      gradient(iz,:) = gradient(1+crop1,:)
    enddo
    do iz=vel_nz-crop1+1,vel_nz
      gradient(iz,:) = gradient(vel_nz-crop1,:)
    enddo

    injected = 0.
    do ishot=1,sou_nx
      injected(:,sou_ox+1+(ishot-1)*sou_dx) = rands_sou(ishot)*wavelet
    end do ! end of shot loop
    gradient_data = 0.
    call c_model2(sou_z, injected, gradient, rec_z, rec_ox, gradient_data)

    alfa = - sum( dprod( gradient_data, data_diff)) / sum( dprod( gradient_data, gradient_data))

    if(isnan(alfa)) then
      write(0,*) "alfa is NaN, will exit the program"
      exit
    else
      ! Output model
      write(0,*) "alfa", alfa
      refl = refl + alfa * gradient
      call sep_write(refl)
      counter = counter + 1
      call to_history("n3", counter)

    endif
  end do ! end of iteration loop

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
!---------------------------------------------------------------------------
!---------------------------------------------------------------------------
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
