#include "cudaErrors.cu"

#define dloc(z,x) (z)+(x)*n1gpu

__constant__ int nt_gpu;
__constant__ int n1gpu;
__constant__ int n2gpu;
__constant__ int nrec_gpu;
__constant__ int crop1_gpu;
__constant__ int crop2_gpu;
__constant__ int n1s_gpu;
__constant__ int n2s_gpu;
__constant__ int wfld_j_gpu;
texture <float, 2, cudaReadModeElementType> vel_ref;
texture <float, 2, cudaReadModeElementType> vel2_ref;
texture <float, 2, cudaReadModeElementType> wt_ref;
cudaChannelFormatDesc cf;
cudaArray *vel_array=0;
cudaArray *vel2_array=0;
cudaArray *wt_array=0;

float *p0_gpu, *p1_gpu, *p2_gpu, *p3_gpu, *source_gpu, *rf_gpu, *wfld1_gpu, *data_gpu;

#include "wave_kernel2d10o.cu"

__global__ void inject_kernel(int z, int it, float *source_gpu, float *p)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=n2gpu) return;
    p[dloc(z,i)] += source_gpu[it+i*nt_gpu];
}

__global__ void inject2_kernel(float *p1, float *rf_gpu, float *p3)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    register int j=blockIdx.y*blockDim.y+threadIdx.y;
    p3[dloc(i,j)] += p1[dloc(i,j)]*rf_gpu[dloc(i,j)];
}


__global__ void collect_kernel(int z, int ox, int it, float *data_gpu, float *p)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=nrec_gpu) return;
    data_gpu[it+(i)*nt_gpu] = p[dloc(z,i+ox)];
}


__global__ void collect2_kernel(float *wfld_gpu, float *p)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    register int j=blockIdx.y*blockDim.y+threadIdx.y;
    if(i>=n1s_gpu || j>=n2s_gpu) return;
    wfld_gpu[i+(j)*n1s_gpu] = p[dloc(i+crop1_gpu,j+crop2_gpu)];
}

extern "C" void gpu_init(int *nt, float *dt, int *n1, int *n2, float *d1, float *d2, int *gpu_id)
{

    cudaSetDevice(*gpu_id);

    cuda_call(cudaMalloc(    (void**) &p0_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &p1_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &source_gpu,  (*nt)*(*n2)*sizeof(float)));

    //bind texture reference to array
    cf = cudaCreateChannelDesc<float>();
    cuda_call(cudaMallocArray(&vel_array, &cf, *n1, *n2));
    cuda_call(cudaMallocArray(&vel2_array, &cf, *n1, *n2));
    cuda_call(cudaMallocArray(&wt_array, &cf, *n1, *n2));

    cuda_call(cudaMemcpyToSymbol(n1gpu, n1, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n2gpu, n2, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(nt_gpu,  nt,   sizeof(int)));

    float coeffs_cpu[COEFFS_SIZE];
    coeffs_cpu[CZ1]=3.333333333*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX1]=3.333333333*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[CZ2]=-.4761904762*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX2]=-.4761904762*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[CZ3]=.0793650794*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX3]=.0793650794*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[CZ4]=.0099206349*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX4]=.0099206349*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[CZ5]=.0006349206*(*dt)*(*dt)/(2.*(*d1)*(*d1));
    coeffs_cpu[CX5]=.0006349206*(*dt)*(*dt)/(2.*(*d2)*(*d2));
    coeffs_cpu[C0]=coeffs_cpu[CZ5]+coeffs_cpu[CX5];
    coeffs_cpu[C0]+=coeffs_cpu[CZ4]+coeffs_cpu[CX4];
    coeffs_cpu[C0]+=coeffs_cpu[CZ3]+coeffs_cpu[CX3];
    coeffs_cpu[C0]+=coeffs_cpu[CZ2]+coeffs_cpu[CX2];
    coeffs_cpu[C0]+=coeffs_cpu[CZ1]+coeffs_cpu[CX1];
    coeffs_cpu[C0]=-2.*coeffs_cpu[C0];

    cuda_call(cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float)));

}

extern "C" void gpu_init_model(int *nt, int *nrec)
{

    cuda_call(cudaMalloc(    (void**) &data_gpu,  (*nt)*(*nrec)*sizeof(float)));
    cuda_call(cudaMemcpyToSymbol(nrec_gpu,  nrec,   sizeof(int)));

}

extern "C" void gpu_init_model2(int *n1, int *n2)
{

    cuda_call(cudaMalloc(    (void**) &p2_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &p3_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &rf_gpu,  (*n1)*(*n2)*sizeof(float)));

}

extern "C" void gpu_init_prop(int *n1s, int *n2s, int *crop1, int *crop2, int *wfld_j)
{

    cuda_call(cudaMalloc(    (void**) &wfld1_gpu,  (*n1s)*(*n2s)*sizeof(float)));

    cuda_call(cudaMemcpyToSymbol(crop1_gpu,  crop1,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(crop2_gpu,  crop2,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n1s_gpu,  n1s,   sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n2s_gpu,  n2s,   sizeof(int)));

    cuda_call(cudaMemcpyToSymbol(wfld_j_gpu,  wfld_j,   sizeof(int)));

}


extern "C" void gpu_vel_update(float *cpu_vel, float *cpu_vel2)
{
    int n1, n2;

    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int))); 
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int))); 
    cuda_call(cudaMemcpyToArray(vel_array, 0,0, cpu_vel, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaBindTextureToArray(vel_ref, vel_array));
    cuda_call(cudaMemcpyToArray(vel2_array, 0,0, cpu_vel2, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaBindTextureToArray(vel2_ref, vel2_array));
}

extern "C" void gpu_wt_update(float *cpu_wt)
{
    int n1, n2;

    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int))); 
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int))); 
    cuda_call(cudaMemcpyToArray(wt_array, 0,0, cpu_wt, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaBindTextureToArray(wt_ref, wt_array));
}

extern "C" void gpu_model(int *sou_z, float *source_cpu, int *rec_z, int *rec_ox, float *data)
{
    int nt, n1, n2, nrec;
    float *ptemp;

    cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nrec, nrec_gpu, sizeof(int)));

    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemset(p0_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(data_gpu, 0,nt*nrec*sizeof(float)));

    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);

    /* 11 ACTUAL WAVE PROPAGATION */

    for (int it=0; it<nt; it++)
    {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
        ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
        kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));
        kernel_exec(collect_kernel<<<nrec/80+1,80>>>(*rec_z, *rec_ox, it, data_gpu, p1_gpu));
    }

    cuda_call(cudaMemcpy(data, data_gpu, nt*nrec*sizeof(float), cudaMemcpyDeviceToHost));

}

extern "C" void gpu_model2(int *sou_z, float *source_cpu, float *rf_cpu, int *rec_z, int *rec_ox, float *data)
{
    int nt, n1, n2, nrec;
    float *ptemp;

    cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nrec, nrec_gpu, sizeof(int)));

    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemcpy(rf_gpu, rf_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemset(p0_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p2_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p3_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(data_gpu, 0,nt*nrec*sizeof(float)));

    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);

    /* 11 ACTUAL WAVE PROPAGATION */

    for (int it=0; it<nt; it++)
    {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
        ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
        kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));

        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p2_gpu, p3_gpu, p2_gpu));
        ptemp=p2_gpu; p2_gpu=p3_gpu; p3_gpu=ptemp;
        kernel_exec(inject2_kernel<<<dimGrid, dimBlock>>>(p1_gpu, rf_gpu, p3_gpu));

        kernel_exec(collect_kernel<<<nrec/80+1,80>>>(*rec_z, *rec_ox, it, data_gpu, p3_gpu));
    }

    cuda_call(cudaMemcpy(data, data_gpu, nt*nrec*sizeof(float), cudaMemcpyDeviceToHost));

}

extern "C" void gpu_prop(int *sou_z, float *source_cpu, float *wfld, int *i_start)
{
    int nt, n1, n2, n3, wfld_j;
    int n1s, n2s;
    float *ptemp;

    cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1s, n1s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2s, n2s_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int)));

    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemset(p0_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(wfld1_gpu, 0,n1s*n2s*sizeof(float)));

    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
//    printf( "nbx %d  nby %d \n", nblocksx,nblocksy);
    int nblocksx2=n1s/BLOCK_SIZE+1;
    int nblocksy2=n2s/BLOCK_SIZE+1;
    dim3 dimGrid2(nblocksx2,nblocksy2);
    dim3 dimBlock2(BLOCK_SIZE, BLOCK_SIZE);

    n3=*i_start;

    if(n3 == 0)
    {
        for (int it=0; it<nt; it++)
        {
            kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
            ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
            kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));

            if(it%wfld_j==0)
            {
                kernel_exec(collect2_kernel<<<dimGrid2, dimBlock2>>>(wfld1_gpu, p1_gpu));
                cuda_call(cudaMemcpy(&wfld[n1s*n2s*n3], wfld1_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
                n3++;
            }

        }
    }
    else
    {
        for (int it=nt-1; it>=0; it--)
        {
            kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
            ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
            kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));

            if(it%wfld_j==0)
            {
                kernel_exec(collect2_kernel<<<dimGrid, dimBlock>>>(wfld1_gpu, p1_gpu));
                cuda_call(cudaMemcpy(&wfld[n1s*n2s*n3], wfld1_gpu, n1s*n2s*sizeof(float), cudaMemcpyDeviceToHost));
                n3--;
            }

        }
    }
}
